#include <stdlib.h>
#include <string.h>
#include <SDL.h>
#include <SDL_mixer.h>

#include "def.h"
#include "main.h"
#include "input.h"
#include "device.h"
#include "hardware.h"

#ifdef _WINDOWS
#include "win_dig.h"
#endif


#define SND_BUFSIZE_OVERRIDE		"SNOWPLOUGH_SND_BUFFER_SIZE"

void fill_audio(void *udata, Uint8 *stream, int len);
samp getsample(void);


/* legacy support: */
bool							wave_device_available = FALSE ;

bool initsounddevice(void)
{
	return(TRUE);
}

#define MAX_SOUND_EFFECTS			256 

typedef struct Sound_Effect_Cache_struct {
			char *		fname ;
			Mix_Chunk *	chunk ;
			} Sound_Effect_Cache ;

static Sound_Effect_Cache		sound_effect_cache [ MAX_SOUND_EFFECTS ] ;
static int						sound_effect_cache_size = 0 ;
static Mix_Music *				mix_music = NULL ;
static int						audio_avail = FALSE ;
static int						still_playing = -1 ;
static bool						sound_on = FALSE ;
static bool						shuffled_music = FALSE ;
static bool						music_on = FALSE ;


#define NUM_TRACKS 10

static const char * track_list [ NUM_TRACKS ] =  {
		"bandaid" , "east17" , "elton" , "lauper" ,
		"rea" , "rocking" , "slade" , "snow" ,
		"wham" , "wizzard" } ;

static int			track_number = 0 ;

int Play_Sound_Effect ( const char * wav )
{
	int			i ;
	char		complete_file [ BUFSIZ ] ;

	if (( ! audio_avail )
	|| ( ! sound_on ))
	{
		return -1 ;
	}
	for ( i = 0 ; i < sound_effect_cache_size ; i ++ )
	{
		if ( strcmp ( wav , sound_effect_cache [ i ] . fname ) == 0 )
		{
			/* already played */
			return Mix_PlayChannel ( -1 , sound_effect_cache [ i ] . chunk , 0 ) ;
		}
	}
	if ( sound_effect_cache_size == MAX_SOUND_EFFECTS )
	{
		fprintf ( err , "Too many sound effects: %s\n" , wav ) ;
		exit ( 1 ) ;
	}

	/* not played yet */
	i = sound_effect_cache_size ;
	sound_effect_cache_size ++ ;

	sprintf ( complete_file , "soundfx/%s.wav" , wav ) ;

	if (( NULL != ( sound_effect_cache [ i ] . fname = 
				malloc ( strlen ( wav ) + 1 )))
	&& ( NULL != ( sound_effect_cache [ i ] . chunk =
			Mix_LoadWAV ( complete_file ))))
	{
		fprintf ( err , "Add to cache (%d): %s\n" , i , wav ) ;
		strcpy ( sound_effect_cache [ i ] . fname , wav ) ;
	} else {
		fprintf ( err , "Unable to load effect: %s\n%s\n" , 
					complete_file , Mix_GetError () ) ;
		exit ( 1 ) ;
	}

	sound_effect_cache [ i ] . chunk -> volume = MIX_MAX_VOLUME ;

	return Mix_PlayChannel ( -1 , sound_effect_cache [ i ] . chunk , 0 ) ;
}

int Play_Sound_Effect_Once ( const char * wav )
{
	if ( still_playing >= 0 )
	{
		fprintf ( err , "Repeated play of %s quashed.\n" , wav ) ;
		return still_playing ;
	}
	still_playing = Play_Sound_Effect ( wav ) ;
	return still_playing ;
}

static void Channel_Finished ( int ch )
{
	if ( ch == still_playing )
	{
		fprintf ( err , "%d just finished.\n" , ch ) ;
		still_playing = -1 ;
	}
}






void Stop_Sound_Effect ( int channel )
{
	if ( channel >= 0 )
	{
		Mix_HaltChannel ( channel ) ;
	}
}



void Play_Music ( const char * file )
{
	char		complete_file [ BUFSIZ ] ;

	if ( ! audio_avail )
	{
		return ;
	}
	Stop_Music () ;
	if (( file == NULL )
	|| ( strlen ( file ) == 0 ))
	{
		/* advance once along our list */
		track_number = ( track_number + 1 ) % NUM_TRACKS ;
		file = track_list [ track_number ] ;
		shuffled_music = TRUE ;
	} else {
		shuffled_music = FALSE ;
	}


	sprintf ( complete_file , "music/%s.ogg" , file ) ;

	if (( NULL != ( mix_music = Mix_LoadMUS ( complete_file )))
	&& ( 0 == Mix_PlayMusic ( mix_music , -1 )))
	{
		fprintf ( err , "New music: %s\n" , file ) ;
	} else {
		fprintf ( err , "Unable to load music: %s\n%s\n" , 
				file , Mix_GetError () ) ;
	/*	exit ( 1 ) ;  -- who cares, it works without music too. */
	}
}

void Stop_Music ( void )
{
	if ( ! audio_avail )
	{
		return ;
	}
	shuffled_music = FALSE ;
	if ( mix_music != NULL )
	{
		Mix_FreeMusic ( mix_music ) ;
		mix_music = NULL ;
	}
}


static int Track_Randomise ( const void * a , const void * b )
{
	return ( rand () & 0x100 ) ? -1 : 1 ;
}


bool setsounddevice(int base, int irq, int dma, Uint4 samprate, Uint4 bufsize)
{
	bool result = FALSE;
	int sndbufsize = 1024 ;		/* default sound buffer size */

	/*SDL_AudioSpec wanted;
	wanted.freq = samprate;
	wanted.samples = bufsize;
	wanted.channels = 1;
	wanted.format = AUDIO_U8;
	wanted.userdata = NULL;
	wanted.callback = fill_audio; */

#ifdef _VGL
	restorekeyb();
#endif

	if ( getenv ( SND_BUFSIZE_OVERRIDE ) != NULL )
	{
		sndbufsize = atoi ( getenv ( SND_BUFSIZE_OVERRIDE ) ) ;
		if ( sndbufsize < 512 )
		{
			fprintf(err, "Invalid buffer size: %d\n" , sndbufsize ) ;
			exit ( 0 ) ;
		}
	}


	if ((SDL_Init(SDL_INIT_AUDIO)) >= 0)
		if (Mix_OpenAudio(22050, AUDIO_S16SYS, 2, sndbufsize ) >= 0 )
			result = TRUE;
	if (result == FALSE)
	{
		fprintf(err, "Couldn't open audio: %s\n", SDL_GetError());
	} else {
		audio_avail = TRUE ;
		Mix_ChannelFinished ( Channel_Finished ) ;
	}

	qsort ( & track_list , NUM_TRACKS , sizeof ( const char * ) ,
					Track_Randomise ) ;
	{
		int i ;

		for ( i = 0 ; i < NUM_TRACKS ; i ++ )
		{
			fprintf ( err , "Shuffler: %d  %s\n" , i , track_list [ i ] ) ;
		}
	}
	music_on = sound_on = FALSE ;
	Sound_Toggle () ;
	Music_Toggle () ;


#ifdef _VGL
	initkeyb();
#endif

	return(result);
}


void killsounddevice(void)
{
	Mix_Pause(-1);
	//SDL_PauseAudio(1);
}

void Generic_Intermission ( int accept_keys_ms , int timeout )
{
	int ms ;

	for ( ms = 0 ; ms < timeout ; ms += 20 )
	{
//		fillbuffer();
#ifdef _WINDOWS
		do_windows_events();
#endif
		SDL_Delay(20);	/* Let some CPU time go away */

		akeypressed = 0 ;
		checkkeyb();

		if (( akeypressed != 0 )
		&& ( ms >= accept_keys_ms ))
		{
			break ;
		}
	}
}

void Music_Toggle ( void )
{
	if ( ! audio_avail )
	{
		return ;
	}
	music_on = ! music_on ;
	if ( music_on )
	{
		if ( shuffled_music )
		{
			Play_Music ( "" ) ;	/* next track */
		}
		Mix_VolumeMusic (( MIX_MAX_VOLUME * 2 ) / 4 ) ;
	} else {
		Mix_VolumeMusic ( 0 ) ;
	}
}

void Sound_Toggle ( void )
{
	sound_on = ! sound_on ;
}


